@@ -18,20 +18,28 @@ struct jellyfish_state {
 #endif
 
 
+/* Create PyUnicode object from NUL terminated Py_UCS4 string. */
+static PyObject* unicode_from_ucs4(const Py_UCS4 *str)
+{
+    size_t len = 0;
+    while (str[len]) len++;
+    return PyUnicode_FromKindAndData(PyUnicode_4BYTE_KIND, (void*)str, len);
+}
+
 /* Returns a new reference to a PyString (python < 3) or
  * PyBytes (python >= 3.0).
  *
  * If passed a PyUnicode, the returned object will be NFKD UTF-8.
  * If passed a PyString or PyBytes no conversion is done.
  */
-static INLINE PyObject* normalize(PyObject *mod, const Py_UNICODE *pystr) {
+static INLINE PyObject* normalize(PyObject *mod, PyObject *pystr) {
     PyObject *unicodedata_normalize;
     PyObject *normalized;
     PyObject *utf8;
 
     unicodedata_normalize = GETSTATE(mod)->unicodedata_normalize;
     normalized = PyObject_CallFunction(unicodedata_normalize,
-                                       "su", "NFKD", pystr);
+                                       "sO", "NFKD", pystr);
     if (!normalized) {
         return NULL;
     }
@@ -40,20 +48,35 @@ static INLINE PyObject* normalize(PyObject *mod, const Py_UNICODE *pystr) {
     return utf8;
 }
 
-static PyObject * jellyfish_jaro_winkler_distance(PyObject *self, PyObject *args, PyObject *kw)
+static PyObject * jellyfish_jaro_winkler_similarity(PyObject *self, PyObject *args, PyObject *kw)
 {
-    const Py_UNICODE *s1, *s2;
+    PyObject *u1, *u2;
+    Py_UCS4 *s1, *s2;
     Py_ssize_t len1, len2;
     double result;
     int long_tolerance = 0;
     static char *keywords[] = {"s1", "s2", "long_tolerance", NULL};
 
-    if (!PyArg_ParseTupleAndKeywords(args, kw, "u#u#|i", keywords, &s1, &len1, &s2, &len2, &long_tolerance)) {
+    if (!PyArg_ParseTupleAndKeywords(args, kw, "UU|i", keywords, &u1, &u2, &long_tolerance)) {
         PyErr_SetString(PyExc_TypeError, NO_BYTES_ERR_STR);
         return NULL;
     }
+    s1 = PyUnicode_AsUCS4Copy(u1);
+    if (s1 == NULL) {
+        return NULL;
+    }
+    s2 = PyUnicode_AsUCS4Copy(u2);
+    if (s2 == NULL) {
+        PyMem_Free(s1);
+        return NULL;
+    }
+    len1 = PyUnicode_GET_LENGTH(u1);
+    len2 = PyUnicode_GET_LENGTH(u2);
+
+    result = jaro_winkler_similarity(s1, len1, s2, len2, long_tolerance);
+    PyMem_Free(s1);
+    PyMem_Free(s2);
 
-    result = jaro_winkler_distance(s1, len1, s2, len2, long_tolerance);
     // jaro returns a big negative number on error, don't use
     // 0 here in case there's floating point inaccuracy
     // .. used to use NaN but different compilers (*cough*MSVC*cough)
@@ -66,19 +89,34 @@ static PyObject * jellyfish_jaro_winkler_distance(PyObject *self, PyObject *args
     return Py_BuildValue("d", result);
 }
 
-static PyObject * jellyfish_jaro_distance(PyObject *self, PyObject *args)
+static PyObject * jellyfish_jaro_similarity(PyObject *self, PyObject *args)
 {
-    const Py_UNICODE *s1, *s2;
+    PyObject *u1, *u2;
+    Py_UCS4 *s1, *s2;
     Py_ssize_t len1, len2;
     double result;
 
-    if (!PyArg_ParseTuple(args, "u#u#", &s1, &len1, &s2, &len2)) {
+    if (!PyArg_ParseTuple(args, "UU", &u1, &u2)) {
         PyErr_SetString(PyExc_TypeError, NO_BYTES_ERR_STR);
         return NULL;
     }
+    len1 = PyUnicode_GET_LENGTH(u1);
+    len2 = PyUnicode_GET_LENGTH(u2);
+    s1 = PyUnicode_AsUCS4Copy(u1);
+    if (s1 == NULL) {
+        return NULL;
+    }
+    s2 = PyUnicode_AsUCS4Copy(u2);
+    if (s2 == NULL) {
+        PyMem_Free(s1);
+        return NULL;
+    }
 
-    result = jaro_distance(s1, len1, s2, len2);
-    // see earlier note about jaro_distance return value
+    result = jaro_similarity(s1, len1, s2, len2);
+    PyMem_Free(s1);
+    PyMem_Free(s2);
+
+    // see earlier note about jaro_similarity return value
     if (result < -1) {
         PyErr_NoMemory();
         return NULL;
@@ -89,32 +127,61 @@ static PyObject * jellyfish_jaro_distance(PyObject *self, PyObject *args)
 
 static PyObject * jellyfish_hamming_distance(PyObject *self, PyObject *args)
 {
-    const Py_UNICODE *s1, *s2;
+    PyObject *u1, *u2;
+    Py_UCS4 *s1, *s2;
     Py_ssize_t len1, len2;
     unsigned result;
 
-    if (!PyArg_ParseTuple(args, "u#u#", &s1, &len1, &s2, &len2)) {
+    if (!PyArg_ParseTuple(args, "UU", &u1, &u2)) {
         PyErr_SetString(PyExc_TypeError, NO_BYTES_ERR_STR);
         return NULL;
     }
+    len1 = PyUnicode_GET_LENGTH(u1);
+    len2 = PyUnicode_GET_LENGTH(u2);
+    s1 = PyUnicode_AsUCS4Copy(u1);
+    if (s1 == NULL) {
+        return NULL;
+    }
+    s2 = PyUnicode_AsUCS4Copy(u2);
+    if (s2 == NULL) {
+        PyMem_Free(s1);
+        return NULL;
+    }
 
     result = hamming_distance(s1, len1, s2, len2);
+    PyMem_Free(s1);
+    PyMem_Free(s2);
 
     return Py_BuildValue("I", result);
 }
 
 static PyObject* jellyfish_levenshtein_distance(PyObject *self, PyObject *args)
 {
-    const Py_UNICODE *s1, *s2;
+    PyObject *u1, *u2;
+    Py_UCS4 *s1, *s2;
     Py_ssize_t len1, len2;
     int result;
 
-    if (!PyArg_ParseTuple(args, "u#u#", &s1, &len1, &s2, &len2)) {
+    if (!PyArg_ParseTuple(args, "UU", &u1, &u2)) {
         PyErr_SetString(PyExc_TypeError, NO_BYTES_ERR_STR);
         return NULL;
     }
+    len1 = PyUnicode_GET_LENGTH(u1);
+    len2 = PyUnicode_GET_LENGTH(u2);
+    s1 = PyUnicode_AsUCS4Copy(u1);
+    if (s1 == NULL) {
+        return NULL;
+    }
+    s2 = PyUnicode_AsUCS4Copy(u2);
+    if (s2 == NULL) {
+        PyMem_Free(s1);
+        return NULL;
+    }
 
     result = levenshtein_distance(s1, len1, s2, len2);
+    PyMem_Free(s1);
+    PyMem_Free(s2);
+
     if (result == -1) {
         // levenshtein_distance only returns failure code (-1) on
         // failed malloc
@@ -129,16 +196,30 @@ static PyObject* jellyfish_levenshtein_distance(PyObject *self, PyObject *args)
 static PyObject* jellyfish_damerau_levenshtein_distance(PyObject *self,
                                                         PyObject *args)
 {
-    Py_UNICODE *s1, *s2;
+    PyObject *u1, *u2;
+    Py_UCS4 *s1, *s2;
     Py_ssize_t len1, len2;
     int result;
 
-    if (!PyArg_ParseTuple(args, "u#u#", &s1, &len1, &s2, &len2)) {
+    if (!PyArg_ParseTuple(args, "UU", &u1, &u2)) {
         PyErr_SetString(PyExc_TypeError, NO_BYTES_ERR_STR);
         return NULL;
     }
+    len1 = PyUnicode_GET_LENGTH(u1);
+    len2 = PyUnicode_GET_LENGTH(u2);
+    s1 = PyUnicode_AsUCS4Copy(u1);
+    if (s1 == NULL) {
+        return NULL;
+    }
+    s2 = PyUnicode_AsUCS4Copy(u2);
+    if (s2 == NULL) {
+        PyMem_Free(s1);
+        return NULL;
+    }
 
     result = damerau_levenshtein_distance(s1, s2, len1, len2);
+    PyMem_Free(s1);
+    PyMem_Free(s2);
     if (result == -1) {
         PyErr_NoMemory();
         return NULL;
@@ -148,13 +229,12 @@ static PyObject* jellyfish_damerau_levenshtein_distance(PyObject *self,
 
 static PyObject* jellyfish_soundex(PyObject *self, PyObject *args)
 {
-    const Py_UNICODE *str;
-    Py_ssize_t len;
+    PyObject *str;
     PyObject *normalized;
     PyObject* ret;
     char *result;
 
-    if (!PyArg_ParseTuple(args, "u#", &str, &len)) {
+    if (!PyArg_ParseTuple(args, "U", &str)) {
         PyErr_SetString(PyExc_TypeError, NO_BYTES_ERR_STR);
         return NULL;
     }
@@ -181,13 +261,12 @@ static PyObject* jellyfish_soundex(PyObject *self, PyObject *args)
 
 static PyObject* jellyfish_metaphone(PyObject *self, PyObject *args)
 {
-    const Py_UNICODE *str;
-    Py_ssize_t len;
+    PyObject *str;
     PyObject *normalized;
     PyObject *ret;
     char *result;
 
-    if (!PyArg_ParseTuple(args, "u#", &str, &len)) {
+    if (!PyArg_ParseTuple(args, "U", &str)) {
         PyErr_SetString(PyExc_TypeError, NO_BYTES_ERR_STR);
         return NULL;
     }
@@ -214,23 +293,33 @@ static PyObject* jellyfish_metaphone(PyObject *self, PyObject *args)
 
 static PyObject* jellyfish_match_rating_codex(PyObject *self, PyObject *args)
 {
-    const Py_UNICODE *str;
+    PyObject *ustr, *ustr_upper;
+    Py_UCS4 *str;
     Py_ssize_t len;
-    Py_UNICODE *result;
+    Py_UCS4 *result;
     PyObject *ret;
 
-    if (!PyArg_ParseTuple(args, "u#", &str, &len)) {
+    if (!PyArg_ParseTuple(args, "U", &ustr)) {
         PyErr_SetString(PyExc_TypeError, NO_BYTES_ERR_STR);
         return NULL;
     }
+    ustr_upper = PyObject_CallMethod(ustr, "upper", NULL);
+    len = PyUnicode_GET_LENGTH(ustr_upper);
+    str = PyUnicode_AsUCS4Copy(ustr_upper);
+    if (str == NULL) {
+        Py_DECREF(ustr_upper);
+        return NULL;
+    }
 
     result = match_rating_codex(str, len);
+    PyMem_Free(str);
+    Py_DECREF(ustr_upper);
     if (!result) {
         PyErr_NoMemory();
         return NULL;
     }
 
-    ret = Py_BuildValue("u", result);
+    ret = unicode_from_ucs4(result);
     free(result);
 
     return ret;
@@ -239,16 +328,38 @@ static PyObject* jellyfish_match_rating_codex(PyObject *self, PyObject *args)
 static PyObject* jellyfish_match_rating_comparison(PyObject *self,
                                                    PyObject *args)
 {
-    const Py_UNICODE *str1, *str2;
+    PyObject *u1, *u2, *u_upper1, *u_upper2;
+    Py_UCS4 *str1, *str2;
     Py_ssize_t len1, len2;
     int result;
 
-    if (!PyArg_ParseTuple(args, "u#u#", &str1, &len1, &str2, &len2)) {
+    if (!PyArg_ParseTuple(args, "UU", &u1, &u2)) {
         PyErr_SetString(PyExc_TypeError, NO_BYTES_ERR_STR);
         return NULL;
     }
+    u_upper1 = PyObject_CallMethod(u1, "upper", NULL);
+    u_upper2 = PyObject_CallMethod(u2, "upper", NULL);
+    len1 = PyUnicode_GET_LENGTH(u_upper1);
+    len2 = PyUnicode_GET_LENGTH(u_upper2);
+    str1 = PyUnicode_AsUCS4Copy(u_upper1);
+    if (str1 == NULL) {
+        Py_DECREF(u_upper1);
+        Py_DECREF(u_upper2);
+        return NULL;
+    }
+    str2 = PyUnicode_AsUCS4Copy(u_upper2);
+    if (str2 == NULL) {
+        PyMem_Free(str1);
+        Py_DECREF(u_upper1);
+        Py_DECREF(u_upper2);
+        return NULL;
+    }
 
     result = match_rating_comparison(str1, len1, str2, len2);
+    PyMem_Free(str1);
+    PyMem_Free(str2);
+    Py_DECREF(u_upper1);
+    Py_DECREF(u_upper2);
 
     if (result == -1) {
         Py_RETURN_NONE;
@@ -261,15 +372,21 @@ static PyObject* jellyfish_match_rating_comparison(PyObject *self,
 
 static PyObject* jellyfish_nysiis(PyObject *self, PyObject *args)
 {
-    const Py_UNICODE *str;
-    Py_UNICODE *result;
+    PyObject *ustr;
+    Py_UCS4 *str;
+    Py_UCS4 *result;
     Py_ssize_t len;
     PyObject *ret;
 
-    if (!PyArg_ParseTuple(args, "u#", &str, &len)) {
+    if (!PyArg_ParseTuple(args, "U", &ustr)) {
         PyErr_SetString(PyExc_TypeError, NO_BYTES_ERR_STR);
         return NULL;
     }
+    str = PyUnicode_AsUCS4Copy(ustr);
+    len = PyUnicode_GET_LENGTH(ustr);
+    if (str == NULL) {
+        return NULL;
+    }
 
     result = nysiis(str, len);
     if (!result) {
@@ -277,7 +394,7 @@ static PyObject* jellyfish_nysiis(PyObject *self, PyObject *args)
         return NULL;
     }
 
-    ret = Py_BuildValue("u", result);
+    ret = unicode_from_ucs4(result);
     free(result);
 
     return ret;
@@ -285,17 +402,23 @@ static PyObject* jellyfish_nysiis(PyObject *self, PyObject *args)
 
 static PyObject* jellyfish_porter_stem(PyObject *self, PyObject *args)
 {
-    const Py_UNICODE *str;
+    PyObject *ustr;
+    Py_UCS4 *str;
     Py_ssize_t len;
-    Py_UNICODE *result;
+    Py_UCS4 *result;
     PyObject *ret;
     struct stemmer *z;
     int end;
 
-    if (!PyArg_ParseTuple(args, "u#", &str, &len)) {
+    if (!PyArg_ParseTuple(args, "U", &ustr)) {
         PyErr_SetString(PyExc_TypeError, NO_BYTES_ERR_STR);
         return NULL;
     }
+    str = PyUnicode_AsUCS4Copy(ustr);
+    len = PyUnicode_GET_LENGTH(ustr);
+    if (str == NULL) {
+        return NULL;
+    }
 
     z = create_stemmer();
     if (!z) {
@@ -303,18 +426,18 @@ static PyObject* jellyfish_porter_stem(PyObject *self, PyObject *args)
         return NULL;
     }
 
-    result = malloc((len+1) * sizeof(Py_UNICODE));
+    result = safe_malloc((len+1), sizeof(Py_UCS4));
     if (!result) {
         free_stemmer(z);
         PyErr_NoMemory();
         return NULL;
     }
-    memcpy(result, str, len * sizeof(Py_UNICODE));
+    memcpy(result, str, len * sizeof(Py_UCS4));
 
     end = stem(z, result, len - 1);
     result[end + 1] = '\0';
 
-    ret = Py_BuildValue("u", result);
+    ret = unicode_from_ucs4(result);
 
     free(result);
     free_stemmer(z);
@@ -323,12 +446,12 @@ static PyObject* jellyfish_porter_stem(PyObject *self, PyObject *args)
 }
 
 static PyMethodDef jellyfish_methods[] = {
-    {"jaro_winkler_distance", (PyCFunction)jellyfish_jaro_winkler_distance, METH_VARARGS|METH_KEYWORDS,
-     "jaro_winkler_distance(string1, string2, long_tolerance)\n\n"
+    {"jaro_winkler_similarity", (PyCFunction)jellyfish_jaro_winkler_similarity, METH_VARARGS|METH_KEYWORDS,
+     "jaro_winkler_similarity(string1, string2, long_tolerance)\n\n"
      "Do a Jaro-Winkler string comparison between string1 and string2."},
 
-    {"jaro_distance", jellyfish_jaro_distance, METH_VARARGS,
-     "jaro_distance(string1, string2)\n\n"
+    {"jaro_similarity", jellyfish_jaro_similarity, METH_VARARGS,
+     "jaro_similarity(string1, string2)\n\n"
      "Get a Jaro string distance metric for string1 and string2."},
 
     {"hamming_distance", jellyfish_hamming_distance, METH_VARARGS,
